#ifndef DEQUE_HPP
#define DEQUE_HPP

#include <cstddef>

// Linked list based deque after Sedgewick and Wayne, Algorithms 4th ed, Algorithm 1.3

template <typename T>
class Deque {

private:

    ////////////////////////////////////
    class Node {
    public:
        T item;
        Node *next;
        Node *prev;
        };

public:

    ////////////////////////////////////
    class iterator;
    friend class iterator;
    class iterator {

    public:

        iterator( Node *c );

        iterator& operator++();
        iterator operator++(int);

        T operator*();
        bool operator!=( typename Deque<T>::iterator );

    private:

        Node *curr;
        };

    ////////////////////////////////////

    Deque( void );
    ~Deque( void );

    void push_left( T &item );
    void push_right( T &item );
    T pop_left( void );
    T pop_right( void );

    bool is_empty( void );
    size_t size( void );

    iterator begin( void );
    iterator end( void );


private:

    size_t N;
    Node *first;
    Node *last;

    };



#endif

