// g++ -D STACK_MAIN -std=c++0x Stack.cpp
// g++ -D STACK_MAIN -std=c++11 Stack.cpp


#include "Stack.hpp"

#include <cassert>
#include <cstddef>

#include <iostream>

template <typename T>
Stack<T>::Stack( void ) :
    N (0),
    list (nullptr)
    {
    }
    
template <typename T>
Stack<T>::~Stack( void ) {
    assert( N == 0 );  // Catch the potential memory leak.
    while ( list ) {
        Node * next = list->next;
        delete list;
        list = next;
        }
    }

template <typename T>
void Stack<T>::push( T &item ) {
    Node * node = new Node();
    node->item = item;
    node->next = list;
    list = node;
    N++;
    }

template <typename T>
T Stack<T>::pop( void ) {
    assert( N > 0 );

    T item = list->item;

    Node *old = list;
    list = list->next;
    delete old;
    N--;

    return item;
    }

template <typename T>
bool Stack<T>::is_empty( void ) {
    return N == 0;
    }

template <typename T>
size_t Stack<T>::size( void ) {
    return N;
    }

template <typename T>
typename Stack<T>::iterator Stack<T>::begin( void ) {
    return Stack<T>::iterator( list );
    }

template <typename T>
typename Stack<T>::iterator Stack<T>::end( void ) {
    return Stack<T>::iterator( nullptr );
    }


////////////////////////////////////////////////////////////////////////////////

template <typename T>
Stack<T>::iterator::iterator( Node *c ) :
    curr (c)
    {
    }

template <typename T>
typename Stack<T>::iterator & Stack<T>::iterator::operator++() {
    if ( this->curr != nullptr ) {
        this->curr = this->curr->next;
        }
    return *this;
    }

template <typename T>
typename Stack<T>::iterator Stack<T>::iterator::operator++( int ) {
    auto t = Stack<T>::iterator( *this );

    if ( this->curr != nullptr ) {
        this->curr = this->curr->next;
        }
    return t;
    }

template <typename T>
T Stack<T>::iterator::operator*() {
    return this->curr->item;
    }

template <typename T>
bool Stack<T>::iterator::operator!=( typename Stack<T>::iterator other ) {
    return this->curr != other.curr;
    }

////////////////////////////////////////////////////////////////////////////////


#ifdef EXERCISES

// 1.3.19
// Returns the item if entry was found.  If the entry did not exist, then the returned value is undefined.
template <typename T>
T Stack<T>::delete_bottom( void ) {
    Node *prev, *curr;
    T item;

    prev = nullptr;
    curr = list;
    if ( curr ) {
        while ( curr->next ) {
            prev = curr;
            curr = curr->next;
            }

        item = curr->item;
        delete curr;
        if ( prev )
            prev->next = nullptr;
        else
            list = nullptr;

        N--;
        }

    return item;
    }


// 1.3.20
// Returns the item if entry was found.  If the entry did not exist, then the returned value is undefined.
template <typename T>
T Stack<T>::remove( size_t k ) {
    Node *prev, *curr;
    size_t i=0;
    T item;

    prev = nullptr;
    curr = list;
    while ( curr && (i < k) ) {
        prev = curr;
        curr = curr->next;
        ++i;
        }

    if ( curr &&( i == k) ) {
        item = curr->item;
        if ( prev )
            prev->next = curr->next;
        else
            list = curr->next;;
        delete curr;
        N--;
        }

    return item;
    }

// 1.3.21
// Returns true if the specified key is in the stack.
template <typename T>
bool Stack<T>::find( T key ) {
    Node *curr = list;

    while ( curr ) {
        if ( curr->item == key )
            return true;
        curr = curr->next;
        }

    return false;
    }

#endif


////////////////////////////////////////////////////////////////////////////////

#ifdef STACK_MAIN

#include <iostream>

int main ( int argc, char **argv ) {

    Stack<long> stack;

    long i;
    while ( ! std::cin.eof() ) {
        std::cin >> i;
        if ( std::cin.good() ) 
            stack.push(i);
        }

    std::cout << "Stack has " << stack.size() << " entries." << std::endl;

    for ( auto iter = stack.begin(); iter != stack.end(); ++iter ) {
        std::cout << *iter << std::endl;
        }

#ifdef EXERCISES
    std::cout << "Looking for 3: " << ( stack.find( 3 ) ? "found" : "not found" ) << std::endl;
    std::cout << "Looking for 9999: " << ( stack.find( 9999 ) ? "found" : "not found" ) << std::endl;

    std::cout << "Removing bottom entry: ";
    i = stack.delete_bottom();
    std::cout << i << std::endl;
    std::cout << "Stack has " << stack.size() << " entries." << std::endl;

    for ( auto iter = stack.begin(); iter != stack.end(); ++iter ) {
        std::cout << *iter << std::endl;
        }

    std::cout << "Removing entry 2: ";
    i = stack.remove( 2 );
    std::cout << i << std::endl;
    std::cout << "Stack has " << stack.size() << " entries." << std::endl;

    for ( auto iter = stack.begin(); iter != stack.end(); ++iter ) {
        std::cout << *iter << std::endl;
        }

#endif

    std::cout << "Popping entries..." << std::endl;

    while ( ! stack.is_empty() ) {
        i = stack.pop();
        std::cout << i << std::endl;
        }

    std::cout << "Stack has " << stack.size() << " entries." << std::endl;

    }

#endif
